package gui;	// FIXME: Can't stay here

import java.awt.AWTException;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.MouseInfo;
import java.awt.Robot;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.security.AccessControlException;
import java.util.ArrayList;
import javax.swing.JComponent;
import javax.swing.JWindow;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.ColorUIResource;
import javax.swing.plaf.FontUIResource;

// 1) Fixing the pointer when in 'fixed' mode doesn't work.  Ideally when
// the thumb is dragged in fixed mode the pointer should remain glued to
// the thumb (unmoving on screen) as the track orientates around it.  We use
// java.awt.Robot to do this.  The problem is, while it's simple to get the
// robot itself to ignore 'artificial' mouse moves, AWT/Swing responds to
// them, meaning the bogus mouse move cancels out the original genuine one!
// Widening the thread lock to make the event thread wait until the robot has
// finished (and gobbled all the bogus events itself) creates a deadlock, as
// it seems java.awt.Robot needs the event thread active in order to do its
// job.  Solution: Gawd alone knows!!  :-)
//
// 2) We need to deal with the circumstance where the track does off the edge
// of the screen.  (This isn't a problem in fixed mode, obviously.)  In such a
// circumstance I suggest we relocate the track/thumb/pointer so the track
// fits entirely on screen.  This will visually be a little jarring (as
// everything will jump sideways) but it's probably preferrable to not
// allowing the user to access the full range of the slider track.
//
// 3) When the slider track fits entirely within the containing Swing
// window, it should be drawn as a lightweight Swing component (on one of
// the window's floating panes) rather than an OS window via AWT.
//
// 4) Drawing needs optimising.  The entire track and thumb base don't
// need to be re-drawn each time - we could use a BufferedImage to hold
// pre-rendered images and simply check each time paint() is called to
// ensure the buffered images still match the size of the component.
//
// 5) The visual look of the component needs reviewing.  The unusual shape
// is meant to set it apart from a regular button, but it does look a little
// strange compared to other Swing components.  We need something which is
// compact, yet suggests 'dragable', while also working well within a
// rectangular track area.  (It would be lovely to have a popup track
// which was just a thin scored line - but until Swing has better support for
// transparency at heavyweight window level, this isn't wise!)
//
// 6) The components don't scale well.  It makes no sense to allow massive or
// odd proportions (eg. try maximising the JPopupSliderTest application window)
// which result in absurd fat slider tracks.
//
public class JPopupSlider extends JComponent implements SwingConstants
{	private int value;						// Current value
	private int minimum,maximum;			// Max and min values
	private int orientation=HORIZONTAL;		// Orientation
	private boolean percentMode=false;		// Label is percent
	private boolean fixedMode=false;		// Thumb is fixed, scrollbar moves
	private boolean showMax=false;			// Non % label shows max range

	private int trackScale;					// Pixel size for each position on the track

	private Popup popup;					// Popup 'track' window
	private Point dragStartPoint;			// X/Y when mouse clicked
	private Point dragStartPointAbs;		// Absolute (screen) x/y when clicked
	private int dragStartValue;				// Value when mouse clicked
	private boolean ignoreEvents;			// Used to discard robot generated events

	private final static int INSET=4;		// Top/bottom border size
	private final static int THUMB_HEX_INSET=6; // Hex egde inset

	private ArrayList<ChangeListener> listeners;


	// -----------------------------------------------------------------
	// CONSTRUCTOR
	// -----------------------------------------------------------------
	public JPopupSlider()
	{	this(HORIZONTAL , 0,0,100 , false,false,false);
	}
	public JPopupSlider(int or)
	{	this(or , 0,0,100 , false,false,false);
	}
	public JPopupSlider(int val,int min,int max)
	{	this(HORIZONTAL , val,min,max , false,false,false);
	}
	public JPopupSlider(int or,int val,int min,int max,boolean pc,boolean smax,boolean fix)
	{	orientation=or;						// Horizontal or vertical
		value=val;							// Starting pos of thumb
		minimum=min;  maximum=max;			// Max/min values of track
		percentMode=pc;						// Show label as "n%" or "n"
		showMax=smax;						// Show label as "n/max"
		fixedMode=fix;						// Move track, not thumb
		normalise();

		// FIXME: Find nicer way to scale track
		trackScale=1;
		while((maximum-minimum)*trackScale>300 && trackScale>1)  trackScale--;
		while((maximum-minimum)*trackScale<100 && trackScale<5)  trackScale++;

		// -----Change event listeners
		listeners = new ArrayList<ChangeListener>();

		// -----Init gui (assumes the code constructing us is on the
		// -----Swing event thread, as per guidelines.
		initGUI();
	}

	// -----------------------------------------------------------------
	// This code is run via the constructor, on the Swing thread
	// -----------------------------------------------------------------
	private void initGUI()
	{	// -----Set the font to our resource font
		Font f = (FontUIResource)UIManager.get("Button.font");
		if(f!=null)  setFont(f);

		// -----Parent frame
		Component c = this;
		while(!(c instanceof Frame) && c!=null)  c=c.getParent();
		// -----Popup track for slider to slid on
		popup = new Popup((Frame)c);

		// -----This listener waits for clicks which signal the raising or
		// -----lowering of the track window.
		this.addMouseListener
		(	new MouseAdapter()
			{	public void mousePressed(MouseEvent ev)
				{	// -----Position within component
					dragStartPoint = ev.getPoint();
					// -----Position on screen
					Point p =((Component)ev.getSource()).getLocationOnScreen();
					dragStartPointAbs = new Point(dragStartPoint);
					dragStartPointAbs.translate(p.x,p.y);
					// -----Initial value
					dragStartValue = value;
					// -----Make the track appear
					popup.setVisible(true);
				}
				public void mouseReleased(MouseEvent ev)
				{	popup.setVisible(false);
					dragStartPoint=null;
				}
			}
		);
		// -----When the track popup window appears, it doesn't get the focus.
		// -----Instead, our humble little thumb button retains the focus.  And
		// -----so, this code bounces motion events to the popup window.
		this.addMouseMotionListener
		(	new MouseMotionAdapter()
			{	public void mouseDragged(MouseEvent ev)
				{	Point p = dragStartPoint;
					if(p!=null)
					{	// -----We need to translate to move the co-ords from
						// -----our screen position to the Popup window's
						ev.translatePoint(0-p.x , 0-p.y);
						popup.updateMouseDrag(ev);
						fireStateChanged();
					}
				}
			}
		);
	}

	// -----------------------------------------------------------------
	// Accessors
	// -----------------------------------------------------------------
	public void setValue(int v) { value=v;  normalise();  repaint(); }
	public void setMinimum(int v) { minimum=v;  normalise();  repaint(); }
	public void setMaximum(int v) { maximum=v;  normalise();  repaint(); }

	public int getValue() { return value; }
	public int getMinimum() { return minimum; }
	public int getMaximum() { return maximum; }

	// -----------------------------------------------------------------
	// Sizing (overrides JComponent)
	// -----------------------------------------------------------------
	public Dimension getPreferredSize()
	{	FontMetrics fm=getFontMetrics(getFont());
		int fontH=fm.getMaxAscent()+fm.getMaxDescent();
		int fontW=fm.stringWidth(labelString(maximum));
		return new Dimension(fontW+THUMB_HEX_INSET*2,fontH+INSET*2);
	}
	public Dimension getMinimumSize() { return getPreferredSize(); }
	public Dimension getMaximumSize() { return getPreferredSize(); }

	// -----------------------------------------------------------------
	// Paint the actual component
	// -----------------------------------------------------------------
	public void paintComponent(Graphics g)
	{	drawThumbBase(g);					// Paint the thumb base
		drawThumbLabel(g);					// Add the text label
	}

	// -----------------------------------------------------------------
	// Event listener code
	// -----------------------------------------------------------------
	public void addChangeListener(ChangeListener l)
	{	synchronized(listeners)
		{	if(!listeners.contains(l))  listeners.add(l);
		}
	}
	public void removeChangeListener(ChangeListener l)
	{	synchronized(listeners)
		{	listeners.remove(l);
		}
	}
	public void fireStateChanged()
	{	ChangeListener[] la;
		synchronized(listeners)
		{	la = new ChangeListener[listeners.size()];
			la = listeners.toArray(la);
		}
		ChangeEvent ev = new ChangeEvent(this);
		for(int i=0;i<la.length;i++)
			la[i].stateChanged(ev);
	}

	// -----------------------------------------------------------------
	// Utility drawing methods
	// -----------------------------------------------------------------
	private void drawThumbBase(Graphics g)
	{	Dimension d=getSize();

		int[] polyX = new int[6];
		int[] polyY = new int[6];
		int mid = d.height/2;

		// -----Background fill
		populatePolyXY(0,0,d.width-1,d.height-1,mid , polyX,polyY);
		//g.setColor(getUIColor("Button.shadow",0x77,0x77,0x77));
		//g.fillPolygon(polyX,polyY,6);
		Color c1 = getUIColor("Button.light",0xCC,0xCC,0xCC);
		Color c2 = getUIColor("Button.shadow",0x33,0x33,0x33);
		Graphics2D g2=(Graphics2D)(g.create());
		g2.setPaint(new GradientPaint(0,0,c1 , 0,d.height-1,c2));
		g2.fillPolygon(polyX,polyY,6);

		// -----Dark border
		g.setColor(getUIColor("Button.darkShadow",0x00,0x00,0x00));
		g.drawPolygon(polyX,polyY,6);

		// -----Edge light
		populatePolyXY(1,1,d.width-2,d.height-2,mid , polyX,polyY);
		g.setColor(getUIColor("Button.light",0xCC,0xCC,0xCC));
		drawLines(g,polyX,polyY , 2,1,0,5);
		// -----Edge shadow
		g.setColor(getUIColor("Button.shadow",0x33,0x33,0x33));
		drawLines(g,polyX,polyY , 5,4,3,2);
	}
	// Draws only the thumb text
	private void drawThumbLabel(Graphics g)
	{	Dimension d=getSize();

		String lab = labelString(value);
		FontMetrics fm=getFontMetrics(getFont());
		int textH=fm.getMaxAscent()+fm.getMaxDescent();
		int textW=fm.stringWidth(lab);

		g.setColor(getUIColor("Button.foreground",0x00,0x00,0x00));
		g.drawString(lab,(d.width-textW)/2 , (d.height-textH)/2+fm.getMaxAscent());
	}
	// Draw track
	private void drawTrack(Graphics g)
	{	Dimension d = popup.getSize();

		int x1=0 , y1=0 , x2=d.width-1 , y2=d.height-1;

		// -----Background
		//g.setColor(getUIColor("Button.background",0x77,0x77,0x77));
		//g.fillRect(0,0,d.width,d.height);
		Color c1 = getUIColor("Button.light",0xCC,0xCC,0xCC);
		Color c2 = getUIColor("Button.shadow",0x33,0x33,0x33);
		Graphics2D g2=(Graphics2D)(g.create());
		if(orientation==HORIZONTAL)
			g2.setPaint(new GradientPaint(0,0,c2 , 0,y2,c1));
		else
			g2.setPaint(new GradientPaint(0,0,c2 , x2,0,c1));
		g2.fillRect(0,0,d.width,d.height);


		// -----Dark border
		g.setColor(getUIColor("Button.darkShadow",0x00,0x00,0x00));
		g.drawRect(x1,y1,x2,y2);

		// -----Light and dark egde
		x1++;  y1++;  x2--;  y2--;
		g.setColor(getUIColor("Button.light",0xCC,0xCC,0xCC));
		g.drawLine(x1,y2,x2,y2);  g.drawLine(x2,y2,x2,y1);
		g.setColor(getUIColor("Button.shadow",0x33,0x33,0x33));
		g.drawLine(x1,y2,x1,y1);  g.drawLine(x1,y1,x2,y1);
	}
	// Fetches a colour resource, else provides constructs default
	private ColorUIResource getUIColor(String k,int r,int g,int b)
	{	ColorUIResource cuir = (ColorUIResource)UIManager.get(k);
		return (cuir==null) ? new ColorUIResource(r,g,b) : cuir ;
	}
	// Draw a list of lines with given Graphics, using indexes (points) from x/y arrays
	private void drawLines(Graphics g,int[] polyX,int[] polyY,int... points)
	{	for(int i=0;i<points.length-1;i++)
		{	g.drawLine
			(	polyX[points[i]],polyY[points[i]] ,
				polyX[points[i+1]],polyY[points[i+1]]
			);
		}
	}
	// Populate x/y arrays with points to form a hexagon
	private void populatePolyXY(int x1,int y1,int x2,int y2,int mid , int[] polyX,int[] polyY)
	{	polyX[0]=x1+THUMB_HEX_INSET;	polyY[0]=y1;		// Top left
		polyX[1]=x1;					polyY[1]=mid;		// Middle left
		polyX[2]=x1+THUMB_HEX_INSET;	polyY[2]=y2;		// Bottom left
		polyX[3]=x2-THUMB_HEX_INSET;	polyY[3]=y2;		// Bottom right
		polyX[4]=x2;					polyY[4]=mid;		// Middle right
		polyX[5]=x2-THUMB_HEX_INSET;	polyY[5]=y1;		// Top right
	}
	// Calculate the maximum label String (used to determine component dimensions)
	private String labelString(int v)		// FIX: crude and sometimes inaccurate
	{	StringBuffer sb = new StringBuffer().append(v);
		if(showMax)  sb.append("/").append(maximum);
		if(percentMode)  sb.append("%");
		return sb.toString();
	}
	// Puts important data within valid ranges
	private void normalise()
	{	// -----Swap max/min is necessary
		if(minimum>maximum)
		{	int t=minimum;  minimum=maximum;  maximum=t;
		}
		// -----Contain value within max/min
		if(value<minimum)  value=minimum;
		else if(value>maximum)  value=maximum;
	}


	// *****************************************************************
	// Popup track window.  This window popups up when the main thumb
	// component is clicked.
	// *****************************************************************
	private class Popup extends JWindow
	{	private Point thumbLoc;				// Location of thumb on screen
		private Dimension thumbDim;			// Pixel width and height of thumb
		private Point trackLoc;				// Location of track
		private Dimension trackDim;			// Pixel width/height of track

		private Robot robot;				// Screw around with AWT mouse events
		private RobotThread robotThread;	// Can't use Robot via AWT thread

		// -------------------------------------------------------------
		// CONSTRUCTOR
		// -------------------------------------------------------------
		Popup(Frame f)
		{	super(f);

			// -----If the thumb remains steady, and the track moves around it,
			// -----we need to create a Robot to allow us to send fake mouse
			// -----events.  Set waitForIdle, as we do not wish to receive our
			// -----own bogus events.
			if(fixedMode)
			{	try
				{	// -----This may fail if security settings deny it
					robot = new Robot();		// Danger, danger, Will Robinson!
					robot.setAutoWaitForIdle(true);
					// -----We can't call the Robot from an AWT thread, so kick off
					// -----our own thread and use wait/notify to signal it.
					robotThread = new RobotThread();
					robotThread.setDaemon(true);
					robotThread.start();
				}
				catch(AWTException e)
				{	//fixedMode=false;
				}
				catch(AccessControlException e)
				{	//fixedMode=false;
				}
			}
		}

		// -------------------------------------------------------------
		// DESTRUCTOR
		// -------------------------------------------------------------
		public void finalize()
		{	// -----Kill the robot!
			robot=null;
			robotThread.interrupt();
		}

		// -------------------------------------------------------------
		// Overrides parent setVisible, to perform init before showing
		// -------------------------------------------------------------
		public void setVisible(boolean b)
		{	if(b==true)  initPopup();
			super.setVisible(b);
		}

		// -------------------------------------------------------------
		// paintComponent() no longer works when subclassing direct from
		// JWindow (JWindow isn't a subclass of JComponent)
		// -------------------------------------------------------------
		public void paint(Graphics g)
		{	// -----Paint the track
			drawTrack(g);

			// -----Create a smaller Graphics for the thumb
			Dimension thumbDim = JPopupSlider.this.getSize();
			Graphics g2;
			if(orientation==HORIZONTAL)
				g2=g.create((value-minimum)*trackScale,0 , thumbDim.width,thumbDim.height);
			else
				g2=g.create(0,(value-minimum)*trackScale , thumbDim.width,thumbDim.height);
			g2.setFont(JPopupSlider.this.getFont());
			drawThumbBase(g2);
			drawThumbLabel(g2);
		}

		// -------------------------------------------------------------
		// This is a drag event which has been bounced to us from the main
		// thumb component.  Its co-ords have been adjusted to be relative
		// to the start point of the drag (see class member 'dragStartPoint')
		// so a leftwards drag will result in negative values.
		// -------------------------------------------------------------
		void updateMouseDrag(MouseEvent ev)
		{	// -----Move the thumb
			if(orientation==HORIZONTAL)
				value=dragStartValue+(ev.getX()/trackScale);
			else
				value=dragStartValue+(ev.getY()/trackScale);
			normalise();
			this.repaint();

			// -----Fixed mode?
			if(fixedMode)
			{	// -----Move the track window (against the direction of the thumb)
				if(orientation==HORIZONTAL)
					this.setLocation(thumbLoc.x-((value-minimum)*trackScale) , thumbLoc.y);
				else
					this.setLocation(thumbLoc.x , thumbLoc.y-((value-minimum)*trackScale));
				// -----Wake the robot thread, so it will move the pointer
				synchronized(JPopupSlider.this) { JPopupSlider.this.notify(); }
			}
		}

		// Popup track window so thumb can be slid
		private void initPopup()
		{	// -----The 'thumb' is the button shown when not popped up
			thumbLoc = JPopupSlider.this.getLocationOnScreen();
			thumbDim = JPopupSlider.this.getSize();
			// -----The 'track' is the popup track for the button to slide on
			if(orientation==HORIZONTAL)
			{	trackDim = new Dimension
				(	thumbDim.width+((maximum-minimum)*trackScale) ,
					thumbDim.height
				);
				trackLoc = new Point
				(	thumbLoc.x-(value-minimum)*trackScale ,
					thumbLoc.y
				);
			}
			else
			{	trackDim = new Dimension
				(	thumbDim.width ,
					thumbDim.height+((maximum-minimum)*trackScale)
				);
				trackLoc = new Point
				(	thumbLoc.x ,
					thumbLoc.y-(value-minimum)*trackScale
				);
			}
			setSize(trackDim);  setLocation(trackLoc);
		}

		// *************************************************************
		// Takes the robot activity off of the AWT/Swing thread
		// *************************************************************
		private class RobotThread extends Thread
		{	public void run()
			{	// -----When Popup.robot member is nullified, time to exit.
				while(robot!=null)
				{	try
					{	synchronized(JPopupSlider.this) { JPopupSlider.this.wait(); }
						// FIXME: Uncommenting this line to attempt to move the
						// pointer back to its drag origin
						//robot.mouseMove(dragStartPointAbs.x,dragStartPointAbs.y);
					}catch(InterruptedException e) {}
				}
			}
		}
	}
}
